#include "bound.h"
#include "gillib.h"

UINT_32 CBound::m_uiObjectCount = 0;
UINT_32 CBound::m_uiNewedVertices = 0;
UINT_32 CBound::m_uiAllocatedVertices = 0;
CListInt *CBound::m_plistFreeVertices = NULL;

CListInt *CBound::m_plistCulledVertices = NULL;
CListInt *CBound::m_plistClippedVertices = NULL;

CBound::CBound()
{
  if ( m_uiObjectCount == 0 )
    {
      m_plistFreeVertices = new CListInt;
      ASSERT( m_plistFreeVertices != NULL );
      m_plistFreeVertices->Init();

      m_plistClippedVertices = new CListInt;
      ASSERT( m_plistClippedVertices != NULL );
      m_plistClippedVertices->Init();

      m_plistCulledVertices = new CListInt;
      ASSERT( m_plistCulledVertices != NULL );
      m_plistCulledVertices->Init();
    }
  m_uiObjectCount++;

  Null();
}

CBound::~CBound()
{
  m_uiObjectCount--;
  if ( m_uiObjectCount == 0 )
    {
      // Clean up our free list of Vertices
      ASSERT( m_uiAllocatedVertices == 0 );
      if ( m_plistFreeVertices != NULL )
	{
	  Vertex *pvx;

	  for ( pvx = (Vertex*)m_plistFreeVertices->RemoveFirst();
		pvx != NULL;
		pvx = (Vertex*)m_plistFreeVertices->RemoveFirst() )
	    {
	      delete pvx;
	    }

	  delete m_plistFreeVertices;
	  m_plistFreeVertices = NULL;

	  m_uiNewedVertices = 0;
	}

      if ( m_plistCulledVertices != NULL )
	{
	  ASSERT( m_plistCulledVertices->GetLength() == 0 );
	  delete m_plistCulledVertices;
	  m_plistCulledVertices = NULL;
	}

      if ( m_plistClippedVertices != NULL )
	{
	  ASSERT( m_plistClippedVertices->GetLength() == 0 );
	  delete m_plistClippedVertices;
	  m_plistClippedVertices = NULL;
	}
    }
}

VOID CBound::Null()
{
  m_bEmpty = TRUE;
  m_pptMinMax[bbtMin].Set(INFINITY, INFINITY, INFINITY);
  m_pptMinMax[bbtMax].Set(-INFINITY, -INFINITY, -INFINITY);
}

VOID CBound::Insert(const Point &pt)
{
  UINT_32 i;

  m_bEmpty = FALSE;
  for ( i = 0; i < 3; i++ )
    {
      m_pptMinMax[bbtMin][i] = MIN(m_pptMinMax[bbtMin][i], pt(i));
      m_pptMinMax[bbtMax][i] = MAX(m_pptMinMax[bbtMax][i], pt(i));
    }
}

VOID CBound::Insert(const CBound &bound)
{
  if ( bound.GetEmpty() )
    {
      return;
    }

  m_bEmpty = FALSE;
  Insert(bound.Get(bbtMin));
  Insert(bound.Get(bbtMax));
}

VOID CBound::Insert(const Matrix &m, const CBound &bound)
{
  UINT_32 i;
  Point pt;

  if ( bound.GetEmpty() )
    {
      return;
    }

  m_bEmpty = FALSE;
  for ( i = 0; i < 8; i++ )
    {
      pt.Set(bound.Get((BoundType)((i>>2)&1))(X), 
	     bound.Get((BoundType)((i>>1)&1))(Y), 
	     bound.Get((BoundType)((i>>0)&1))(Z));
      pt = m * pt;
      Insert(pt);
    }  
}

BOOL CBound::InBounds(const Point &pt)
{
  return (m_pptMinMax[bbtMin] <= pt && 
	  m_pptMinMax[bbtMax] >= pt);
}

//////////////////////////////////////////////////////////////////////////////
// START: ASSIGNMENT 4: Bound Culling
// 

// Returns TRUE if pt is in the half space inside the bound defined by the 
// the edge: eType = [Min or Max] and uiCoord = [X or Y]
BOOL CBound::InHalf(CBound::BoundType eType, UINT_32 uiCoord, const Point &pt)
{
  return TRUE;
}

// Returns [bitDisjoint, bitEnclosed, or bitIntersect] based on how the input
// bound interacts with the "this" clip bound
CBound::BoundIntersectionType CBound::Intersect(const CBound &bound)
{
  return bitIntersect;
}

// Returns [bitDisjoint, bitEnclosed, or bitIntersect] based on how the input
// transformed bound interacts with the "this" clip bound
// Note: use the Cohen-Sutherland opcodes to speed this up
CBound::BoundIntersectionType CBound::Intersect(const Matrix &m, 
						const CBound &bound)
{
  return bitIntersect;
}
// 
// END  : ASSIGNMENT 4: Bound Culling
//////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////
// START: ASSIGNMENT 4: Polygon Clipping
//

// Returns the parametric value t in the line equation defined by 
// L = pt0 + t*(pt1 - pt0) of where the line segment intercepts the 
// clip bound on its edge defined by eType = [Min or Max] and 
// uiCoord = [X or Y]
FLOAT CBound::Intercept(BoundType eType, UINT_32 uiCoord, 
			const Point &pt0, const Point &pt1)
{
  return (FLOAT)0.0;
}

// Alters the listVertices to return the polygon contour created by the 
// Sutherland-Hodgman polygon clipping algorithm.
// Any removed Vertices should be deallocated by calling 
// CBound::DeallocateVertex(pvx).
// Any new Vertices should be created by calling CBound::AllocateVertex().
// The list class does not have any convenient alteration methods, so you
// should remove the vertices from the input list and insert them into a
// temporary list on each clip edge operations.
// Remember to move the resulting polygon contour back into the input 
// listVertices before returning.
VOID CBound::ClipPolygon(CListInt &listVertices)
{
  CListInt *plistVertices;
  Vertex *pvx;

  // You may want to use these
  // m_plistCulledVertices  -- temporary list for the vertices which must
  //                           be deallocated
  // m_plistClippedVertices -- temporary list for the contour

  // Here is some dumb code which removes all the vertices from one list, 
  // puts them into another list, and then reverses the process.

  plistVertices = &listVertices;
  ASSERT( m_plistClippedVertices->GetLength() == 0 );
  for ( pvx = (Vertex*)plistVertices->RemoveFirst(); 
	pvx != NULL; 
	pvx = (Vertex*)plistVertices->RemoveFirst() )
    {
      m_plistClippedVertices->InsertLast(pvx, 0);
    }

  ASSERT( plistVertices->GetLength() == 0 );
  for ( pvx = (Vertex*)m_plistClippedVertices->RemoveFirst(); 
	pvx != NULL; 
	pvx = (Vertex*)m_plistClippedVertices->RemoveFirst() )
    {
      plistVertices->InsertLast(pvx, 0);
    }

  ASSERT( m_plistClippedVertices->GetLength() == 0 );
}
// 
// END  : ASSIGNMENT 4: Polygon Clipping
//////////////////////////////////////////////////////////////////////////////

VOID CBound::Render(const Matrix &m)
{
  BoundType i, j, k;
  Point ppppt[2][2][2];
  Point pt;

  if ( GetEmpty() )
    {
      return;
    }

  for ( i = bbtMin; i < bbtSize; i = (BoundType)(i + 1) )
    {
      for ( j = bbtMin; j < bbtSize; j = (BoundType)(j + 1) )
	{
	  for ( k = bbtMin; k < bbtSize; k = (BoundType)(k + 1) )
	    {
	      pt.Set(m_pptMinMax[i][X], 
		     m_pptMinMax[j][Y],
		     m_pptMinMax[k][Z]);
	      ppppt[i][j][k] = m * pt;
	    }
	}
    }

  gilBegin(GL_LINE_LOOP);
  {
    // Left
    gilVertex3fv(ppppt[bbtMin][bbtMin][bbtMin].Get());
    gilVertex3fv(ppppt[bbtMin][bbtMin][bbtMax].Get());
    gilVertex3fv(ppppt[bbtMin][bbtMax][bbtMax].Get());
    gilVertex3fv(ppppt[bbtMin][bbtMax][bbtMin].Get());
  }
  gilEnd();

  gilBegin(GL_LINE_LOOP);
  {
    // Right
    gilVertex3fv(ppppt[bbtMax][bbtMin][bbtMin].Get());
    gilVertex3fv(ppppt[bbtMax][bbtMax][bbtMin].Get());
    gilVertex3fv(ppppt[bbtMax][bbtMax][bbtMax].Get());
    gilVertex3fv(ppppt[bbtMax][bbtMin][bbtMax].Get());
  }
  gilEnd();

  gilBegin(GL_LINE_LOOP);
  {
    // Bottom
    gilVertex3fv(ppppt[bbtMin][bbtMin][bbtMin].Get());
    gilVertex3fv(ppppt[bbtMax][bbtMin][bbtMin].Get());
    gilVertex3fv(ppppt[bbtMax][bbtMin][bbtMax].Get());
    gilVertex3fv(ppppt[bbtMin][bbtMin][bbtMax].Get());
  }
  gilEnd();

  gilBegin(GL_LINE_LOOP);
  {
    // Top
    gilVertex3fv(ppppt[bbtMin][bbtMax][bbtMin].Get());
    gilVertex3fv(ppppt[bbtMin][bbtMax][bbtMax].Get());
    gilVertex3fv(ppppt[bbtMax][bbtMax][bbtMax].Get());
    gilVertex3fv(ppppt[bbtMax][bbtMax][bbtMin].Get());
  }
  gilEnd();

  gilBegin(GL_LINE_LOOP);
  {
    // Back
    gilVertex3fv(ppppt[bbtMin][bbtMin][bbtMin].Get());
    gilVertex3fv(ppppt[bbtMin][bbtMax][bbtMin].Get());
    gilVertex3fv(ppppt[bbtMax][bbtMax][bbtMin].Get());
    gilVertex3fv(ppppt[bbtMax][bbtMin][bbtMin].Get());
  }
  gilEnd();

  gilBegin(GL_LINE_LOOP);
  {
    // Front
    gilVertex3fv(ppppt[bbtMin][bbtMin][bbtMax].Get());
    gilVertex3fv(ppppt[bbtMax][bbtMin][bbtMax].Get());
    gilVertex3fv(ppppt[bbtMax][bbtMax][bbtMax].Get());
    gilVertex3fv(ppppt[bbtMin][bbtMax][bbtMax].Get());
  }
  gilEnd();
}

VOID CBound::Render(const Matrix &mGvp_proj, const Matrix &mGproj_obj)
{
  BoundType i, j, k;
  Point ppppt[2][2][2];
  Point pt;
  Matrix mGvp_obj;

  if ( GetEmpty() )
    {
      return;
    }

  mGvp_obj = mGvp_proj * mGproj_obj;

  for ( i = bbtMin; i < bbtSize; i = (BoundType)(i + 1) )
    {
      for ( j = bbtMin; j < bbtSize; j = (BoundType)(j + 1) )
	{
	  for ( k = bbtMin; k < bbtSize; k = (BoundType)(k + 1) )
	    {
	      pt.Set(m_pptMinMax[i][X], 
		     m_pptMinMax[j][Y],
		     m_pptMinMax[k][Z]);
	      ppppt[i][j][k] = mGvp_obj * pt;
	      ppppt[i][j][k].Homogenize();
	    }
	}
    }

  gilBegin(GL_LINE_LOOP);
  {
    // Left
    gilVertex3fv(ppppt[bbtMin][bbtMin][bbtMin].Get());
    gilVertex3fv(ppppt[bbtMin][bbtMin][bbtMax].Get());
    gilVertex3fv(ppppt[bbtMin][bbtMax][bbtMax].Get());
    gilVertex3fv(ppppt[bbtMin][bbtMax][bbtMin].Get());
  }
  gilEnd();

  gilBegin(GL_LINE_LOOP);
  {
    // Right
    gilVertex3fv(ppppt[bbtMax][bbtMin][bbtMin].Get());
    gilVertex3fv(ppppt[bbtMax][bbtMax][bbtMin].Get());
    gilVertex3fv(ppppt[bbtMax][bbtMax][bbtMax].Get());
    gilVertex3fv(ppppt[bbtMax][bbtMin][bbtMax].Get());
  }
  gilEnd();

  gilBegin(GL_LINE_LOOP);
  {
    // Bottom
    gilVertex3fv(ppppt[bbtMin][bbtMin][bbtMin].Get());
    gilVertex3fv(ppppt[bbtMax][bbtMin][bbtMin].Get());
    gilVertex3fv(ppppt[bbtMax][bbtMin][bbtMax].Get());
    gilVertex3fv(ppppt[bbtMin][bbtMin][bbtMax].Get());
  }
  gilEnd();

  gilBegin(GL_LINE_LOOP);
  {
    // Top
    gilVertex3fv(ppppt[bbtMin][bbtMax][bbtMin].Get());
    gilVertex3fv(ppppt[bbtMin][bbtMax][bbtMax].Get());
    gilVertex3fv(ppppt[bbtMax][bbtMax][bbtMax].Get());
    gilVertex3fv(ppppt[bbtMax][bbtMax][bbtMin].Get());
  }
  gilEnd();

  gilBegin(GL_LINE_LOOP);
  {
    // Back
    gilVertex3fv(ppppt[bbtMin][bbtMin][bbtMin].Get());
    gilVertex3fv(ppppt[bbtMin][bbtMax][bbtMin].Get());
    gilVertex3fv(ppppt[bbtMax][bbtMax][bbtMin].Get());
    gilVertex3fv(ppppt[bbtMax][bbtMin][bbtMin].Get());
  }
  gilEnd();

  gilBegin(GL_LINE_LOOP);
  {
    // Front
    gilVertex3fv(ppppt[bbtMin][bbtMin][bbtMax].Get());
    gilVertex3fv(ppppt[bbtMax][bbtMin][bbtMax].Get());
    gilVertex3fv(ppppt[bbtMax][bbtMax][bbtMax].Get());
    gilVertex3fv(ppppt[bbtMin][bbtMax][bbtMax].Get());
  }
  gilEnd();
}

VOID CBound::Display()
{
  printf("Bound:\n");
  m_pptMinMax[bbtMin].Display();
  m_pptMinMax[bbtMax].Display();
}

Vertex *CBound::AllocateVertex()
{
  Vertex *ppt;

  ASSERT( m_plistFreeVertices != NULL );

  ppt = (Vertex*)m_plistFreeVertices->RemoveFirst();
  if ( ppt == NULL )
    {
      ASSERT( m_uiNewedVertices == m_uiAllocatedVertices );
      m_uiNewedVertices++;
      ppt = new Vertex;
      ASSERT( ppt != NULL );
    }
  ASSERT( m_uiAllocatedVertices < m_uiNewedVertices );

  m_uiAllocatedVertices++;

  return ppt;
}

VOID CBound::DeallocateVertex(Vertex *ppt)
{
  ASSERT( ppt != NULL );
  ASSERT( m_plistFreeVertices != NULL );
  
  ASSERT( m_uiAllocatedVertices > 0 );

  m_plistFreeVertices->InsertFirst(ppt, 0);
  m_uiAllocatedVertices--;
  ASSERT( m_uiAllocatedVertices < m_uiNewedVertices );
}
