#ifndef __BOUND_H__
#define __BOUND_H__

#include "point.h"
#include "matrix.h"
#include "vertex.h"
#include "datalib.h"

//////////////////////////////////////////////////////////////////////////////
// CBound Class
//

class CBound
{
//////////////////////////////////////////////////
// Enumerated Types
//
public:
//////////////////////////////////////////////////////////////////////////////
// START: ASSIGNMENT 4: Bound Culling Intersection Test Return Values
// 
enum BoundIntersectionType {bitDisjoint, bitEnclosed, bitIntersect, bitEnveloping,
			    bitSize};
// 
// END  : ASSIGNMENT 4: Bound Culling Intersection Test Return Values
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
// START: ASSIGNMENT 4: Bound Box Extents Type [Min, Max]
// 
enum BoundType {bbtMin, bbtMax, bbtSize};
// 
// END  : ASSIGNMENT 4: Bound Box Extents Type [Min, Max]
//////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////
// Initialization
//
public:
  CBound();
  ~CBound();

//////////////////////////////////////////////////
// Interface
//
public:
  VOID Null();
  VOID Insert(const Point &pt);
  VOID Insert(const CBound &bound);
  VOID Insert(const Matrix &m, const CBound &bound);
  BOOL InBounds(const Point &pt);
//////////////////////////////////////////////////////////////////////////////
// START: ASSIGNMENT 4: Bound Culling
// 
  // Is the pt in a Half Space
  BOOL InHalf(BoundType eType, UINT_32 uiCoord, const Point &pt);
  // Axis aligned bounding box intersection with axis aligned bounding box
  BoundIntersectionType Intersect(const CBound &bound);
  // Axis aligned bounding box intersection with transformed bounding box
  BoundIntersectionType Intersect(const Matrix &m,
				  const CBound &bound);
// 
// END  : ASSIGNMENT 4: Bound Culling
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
// START: ASSIGNMENT 4: Polygon Clipping
// 
  // Parametric line segment intersection with axis aligned bounding box
  FLOAT Intercept(BoundType eType, UINT_32 uiCoord, 
		  const Point &pt0, const Point &pt1);
  // Sutherland-Hodgman polygon clipping algorithm
  VOID ClipPolygon(CListInt &listVertices);
// 
// END  : ASSIGNMENT 4: Polygon Clipping
//////////////////////////////////////////////////////////////////////////////

public:
  inline BOOL GetEmpty() const;
  inline const Point &Get(BoundType bbt) const;
  inline Point GetCenter();
  inline FLOAT GetRadius();

//////////////////////////////////////////////////
// Rendering
//
public:
  VOID Render(const Matrix &m);
  VOID Render(const Matrix &mGvp_proj, const Matrix &mGproj_obj);

//////////////////////////////////////////////////
// Debugging
//
public:
  VOID Display();


//////////////////////////////////////////////////
// Class Methods
//
public:
//////////////////////////////////////////////////////////////////////////////
// START: ASSIGNMENT 4: Fast Vertex Allocation Class Methods
// 
  static Vertex *AllocateVertex();
  static VOID DeallocateVertex(Vertex *pvx);
// 
// END  : ASSIGNMENT 4: Fast Vertex Allocation Class Methods
//////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////
// Class Variables
//
protected:
  static UINT_32 m_uiObjectCount;
  static UINT_32 m_uiNewedVertices;
  static UINT_32 m_uiAllocatedVertices;
  static CListInt *m_plistFreeVertices;

//////////////////////////////////////////////////////////////////////////////
// START: ASSIGNMENT 4: Lists which may be helpful in the PolygonClip()
// 
  // A list of Vertices which need to be deallocated
  static CListInt *m_plistCulledVertices;  
  // A list for temporarily storing the polygon contour
  static CListInt *m_plistClippedVertices;
// 
// END  : ASSIGNMENT 4: Lists which may be helpful in the PolygonClip()
//////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////
// Member Variables
//
protected:
  BOOL m_bEmpty;
  Point m_pptMinMax[bbtSize];
};

//////////////////////////////////////////////////////////////////////
// External Operators
//
inline BOOL operator==(CBound &bound0, CBound &bound1);
inline BOOL operator!=(CBound &bound0, CBound &bound1);

//////////////////////////////////////////////////////////////////////
// Inlined Functions
//

inline BOOL CBound::GetEmpty() const
{
  return m_bEmpty;
}

const Point &CBound::Get(BoundType bbt) const
{
  ASSERT( bbt < bbtSize );
  
  return m_pptMinMax[bbt];
}

Point CBound::GetCenter()
{
  return m_pptMinMax[bbtMin] + ((m_pptMinMax[bbtMax] - m_pptMinMax[bbtMin])*0.5);
}

FLOAT CBound::GetRadius()
{
  return ((m_pptMinMax[bbtMax] - m_pptMinMax[bbtMin]).Magnitude()) * (FLOAT)0.5;
}

inline BOOL operator==(CBound &bound0, CBound &bound1)
{
  return ( (bound0.Get(CBound::bbtMin) == 
	    bound1.Get(CBound::bbtMin)) &&
	   (bound0.Get(CBound::bbtMax) == 
	    bound1.Get(CBound::bbtMax)) );
}

inline BOOL operator!=(CBound &bound0, CBound &bound1)
{
  return !(bound0 == bound1);
}

#endif // __BOUND_H__
