#include "slidepoint.h"
#include "slideworld.h"
#include "slidesurface.h"

//////////////////////////////////////////////////////////////////////////////
// CSLIDEPoint Class
//

CSLIDEPoint::CSLIDEPoint()
{
  m_pcSurfaceID = NULL;
  m_pCSLIDESurface = NULL;

  m_pt.Null();
  m_vNormal.Null();
  m_ptTexture.Null();
}

CSLIDEPoint::~CSLIDEPoint()
{
}

VOID CSLIDEPoint::Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, 
		       CHAR *pcID, 
		       DynPoint *pdynpt)
{
  ASSERT( pdynpt != NULL );

  CSLIDEPrimitive::Init(pcFile, uiLine, uiColumn, pcID);

  m_pcSurfaceID = gcppcSLIDEReservedWords[sliderwtSLF_INHERIT];
  m_pCSLIDESurface = NULL;

  m_pt.SetAll(pdynpt);
  m_vNormal.Null();
  m_ptTexture.Null();

  SetBoundDynamic(FALSE);
}

VOID CSLIDEPoint::Uninit()
{
  CSLIDEPrimitive::Uninit();
}

BOOL CSLIDEPoint::Initialized()
{
  if ( !CSLIDEPrimitive::Initialized() )
    {
      return FALSE;
    }

  return m_pcSurfaceID != NULL; 
}

BOOL CSLIDEPoint::CheckSemantics(CSLIDEWorld *pCSLIDEWorld)
{
  BOOL bCorrect;

  if ( !CSLIDEPrimitive::CheckSemantics(pCSLIDEWorld) )
    {
      return FALSE;
    }

  bCorrect = TRUE;

  // Check Surface Reference

  m_pCSLIDESurface = pCSLIDEWorld->PeekSurface(m_pcSurfaceID);
  if ( m_pCSLIDESurface == NULL )
    {
      HandleError("Undefined %s \"%s\"", 
		  gcppcSLIDEReservedWords[sliderwtSurface],
		  m_pcSurfaceID);
      bCorrect = FALSE;
    }

  return bCorrect;
}

BOOL CSLIDEPoint::Valid()
{
  if ( !CSLIDEPrimitive::Valid() )
    {
      return FALSE;
    }

  return m_pt[W] != 0.0;
}

VOID CSLIDEPoint::InsertIntoBound(CBound *pCBound)
{
  ASSERT( pCBound != NULL );

  pCBound->Insert(m_pt);
}

VOID CSLIDEPoint::CheckDynamic()
{
  SetDynamic(m_pt.GetDynamic() ||
	     m_vNormal.GetDynamic() ||
	     m_ptTexture.GetDynamic());
  SetBoundDynamic(m_pt.GetDynamic());
}

VOID CSLIDEPoint::UpdateState()
{
  m_pt.Update();

  if ( m_vNormal.GetDynamic() )
    {
      m_vNormal.Update();
      m_vNormal.CompletePlaneEquation(m_pt);
    }
  m_ptTexture.Update();
}

VOID CSLIDEPoint::WriteSLIDE(FILE *fpOutputFile)
{
  Vector v;

  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "%s %s ( %.3f %.3f %.3f )\n", 
	  gcppcSLIDEReservedWords[sliderwtPoint], m_pcID, 
	  m_pt[X], m_pt[Y], m_pt[Z]);

  if ( m_pcSurfaceID != gcppcSLIDEReservedWords[sliderwtSLF_INHERIT] )
    {
      fprintf(fpOutputFile, "  %s %s\n",
	      gcppcSLIDEReservedWords[sliderwtSurface], 
	      m_pcSurfaceID);
    } 
  if ( m_vNormal.Magnitude() != 0.0 )
    {
      fprintf(fpOutputFile, "  %s ( %.3f %.3f %.3f )\n",
	      gcppcSLIDEReservedWords[sliderwtNormal], 
	      m_vNormal[X], m_vNormal[Y], m_vNormal[Z]); 
    }
  v = m_ptTexture;
  if ( v.Magnitude() != 0.0 )
    {
      fprintf(fpOutputFile, "  %s ( %.3f %.3f %.3f )\n",
	      gcppcSLIDEReservedWords[sliderwtTexture], 
	      m_ptTexture[0], m_ptTexture[1], m_ptTexture[2]); 
    }

  fprintf(fpOutputFile, "%s\n", 
	  gcppcSLIDEReservedWords[sliderwtEndPoint]); 
}
